from datetime import timedelta
from functools import partial
from typing import Any, Callable, List, Sequence, Set, Tuple, cast

import hypothesis
import hypothesis.extra.numpy as st_np
import hypothesis.strategies as st
import torch
from hypothesis import example, given

import interp.circuit.testing.strategies as st_c
import rust_circuit as rc
from interp.circuit.circuit_compiler.util import TorchAxisIndex
from interp.circuit.testing.topdown_circuit import CircuitProbs as CP
from interp.circuit.testing.utils import mark_not_interesting_if, target_diff
from rust_circuit import Add
from rust_circuit import Add as rAdd
from rust_circuit import Array
from rust_circuit import Circuit
from rust_circuit import Circuit as rCircuit
from rust_circuit import Concat
from rust_circuit import Einsum
from rust_circuit import Einsum as rEinsum
from rust_circuit import Index, Matcher, Parser, Rearrange
from rust_circuit import Scalar
from rust_circuit import Scalar as rScalar
from rust_circuit import Tag, apply_in_traversal, new_traversal, replace_outside_traversal_symbols, simp
from rust_circuit.algebric_rewrite import (
    ImpossibleRewriteError,
    RewriteError,
    distribute_old,
    drop_from_add,
    drop_mul_ones,
    drop_zero_add,
    einsum_flatten_bans_noop,
    eliminate_zeros,
    explicit_reduce,
    explicit_squeeze,
    extract_add,
    extract_add_by_match,
    extract_input_diags,
    extract_output_diags,
    factor_add_of_mul_to_mul_of_add,
    flatten_adds,
    fuse_einsum_permute,
    fuse_einsum_rearrange,
    fuse_einsum_single,
    fuse_permute_einsum,
    fuse_rearrange,
    fuse_single_einsum,
    multiply_axes_by_identity,
    nested_einsum_axes_permute,
    nested_einsum_permute_dups_to_eq,
    permute_to_einsum,
    pull_through_add_concat,
    pull_through_add_index,
    pull_through_add_rearrange,
    push_down_permute_via_einsum,
    rearrange_of_const_to_const,
    remove_empty_einsum,
    remove_noop_rearrange,
    remove_single_concat,
    remove_trivial_index,
    split_einsum_concat,
    split_to_concat_for_batch,
    zero_empty_add,
)

from .util import dedup_with_order, st_Circuit

P = Parser()
po = rc.PrintOptions(tensor_index_literal=True, bijection=False)


def is_zero(c: Circuit):
    return c.is_scalar() and c.cast_scalar().is_zero()


@st.composite
def st_index_groups(draw: st.DrawFn, length: int) -> List[List[int]]:
    # Generate index groups. Subtract from the available positions at every
    # step. Stop when we draw a False boolean or when we have no space left.
    available_space: List[Tuple[int, int]] = [(0, length)]
    index_groups: List[List[int]] = []
    first_iter = True

    # While there is space left, or we may randomly stop
    while first_iter or (len(available_space) > 0 and draw(st.booleans())):
        first_iter = False

        # Choose an available space
        space_idx = draw(st.integers(0, len(available_space) - 1))
        i_start, i_end = available_space.pop(space_idx)

        # Generate group indices from the available space
        group_start = draw(st.integers(i_start, i_end - 1))
        group_end = draw(st.integers(i_start, i_end - 1))
        group_start, group_end = sorted([group_start, group_end])
        group_end += 1

        # Add back the remaining space
        if i_start != group_start:
            available_space.append((i_start, group_start))
        if i_end != group_end:
            available_space.append((group_end, i_end))

        # Put our group in the generated groups, as a collection of contiguous indices
        index_groups.append(list(range(group_start, group_end)))
    return index_groups


@st.composite
def st_index_groups_and_circuit(
    draw: st.DrawFn, must_be_explicitly_computable: bool = True
) -> Tuple[List[List[int]], int, rCircuit]:
    st_shapes = st_np.array_shapes(min_dims=1, max_dims=4, max_side=8)
    st_circuit = st_Circuit(
        st_shapes,
        must_be_explicitly_computable=must_be_explicitly_computable,
        max_growth_steps=10,
        probs_default=CP.kw(all=1),
        probs_per_depth=[
            CP.kw(all=0, Einsum=(1, dict(min_n_children=1))),
            CP.kw(all=0, Concat=40),
        ],
    )
    circuit = draw(st_circuit)
    assert isinstance(circuit, Einsum) and len(circuit.args) > 0

    concat_idxs = [i for i, c in enumerate(circuit.children) if isinstance(c, Concat)]
    hypothesis.assume(len(concat_idxs) > 0, "no concat indices")
    which = concat_idxs[draw(st.integers(min_value=0, max_value=len(concat_idxs) - 1))]

    index_groups_target: Concat = circuit.children[which].cast_concat()
    index_groups = draw(st_index_groups(index_groups_target.num_children))
    return index_groups, which, circuit


@hypothesis.settings(
    max_examples=10,
    suppress_health_check=[hypothesis.HealthCheck.filter_too_much, hypothesis.HealthCheck.too_slow],
    deadline=timedelta(seconds=4),
)
@given(groups_and_circuit=st_index_groups_and_circuit())
@mark_not_interesting_if(NotImplementedError)
def test_auto_split_einsum_concat(groups_and_circuit):
    index_groups, which, circuit = groups_and_circuit
    check_transform_equality(partial(split_einsum_concat, element_index=which, index_groups=index_groups), circuit)


@mark_not_interesting_if(RewriteError)
# use Any for `fn` type to avoid errors
def check_transform_equality(
    fn: Callable[[Any], rCircuit], circuit: Circuit, sample: bool = False, f64: bool = True
) -> Circuit:
    """
    Checks that when we apply the rewrite `fn` to `circuit` the evaluated circuit has the same value

    ..note:: we make sure that enough cases for each rewrite are tested by marking the test example as invalid if the
        rewrite noops.

      Thus, we enforce that rewrites have to raise :class:`NoopRewriteError` whenever they would noop (or sometimes just
      :class:`NotImplementedError`. This is true for all rewrites *except* :func:`rearrange_muls`. In that case, many successful rearranges
      would leave the tree of Einsums equal, and so this doesn't make sense.
    """
    if f64:
        circuit = rc.cast_circuit(circuit, rc.TorchDeviceDtypeOp(dtype="float64"))
    trans = fn(circuit)

    assert trans.shape == circuit.shape
    assert trans != circuit, "The rewrite should raise a RewriteError instead of nooping."

    if sample:
        sampler = rc.Sampler(rc.RunDiscreteVarAllSpec.create_full_from_circuits(trans, circuit))
        trans_eval = sampler(trans)
        circuit_eval = sampler(circuit)
    else:
        trans_eval = trans
        circuit_eval = circuit

    hypothesis.note(circuit_eval.repr(po))
    hypothesis.note(trans_eval.repr(po))
    actual, expected = (trans_eval.evaluate(), circuit_eval.evaluate())
    target_diff(actual.type(expected.dtype), expected)  # , evaluator=evaluator)
    return trans


@hypothesis.settings(deadline=None)
@given(
    circuit=st_Circuit(
        st_np.array_shapes(min_dims=0),
        must_be_explicitly_computable=True,
        max_growth_steps=10,
        # Sample more relevant nodes at depth 0 and more zeros at depth 1
        probs_default=CP.kw(all=1, Scalar=(5, dict(values=[0.0], also_any=True))),
        probs_per_depth=[CP.kw(all=0, Add=10, Einsum=10, Rearrange=10, Index=10)],
    )
)
def test_eliminate_zeros(circuit):
    """
    Check that if the top level is a :class:`Einsum` or :class:`Rearrange` or :class:`Index` or :class:`Add`, it has been
    replaced by a zero (and the shape is still the same). Also exact float equality.
    """
    transformed_circuit = check_transform_equality(eliminate_zeros, circuit)

    assert not any(is_zero(c) for c in transformed_circuit.children)


@given(
    circuit=st_Circuit(
        st_np.array_shapes(min_dims=0),
        must_be_explicitly_computable=True,
        max_growth_steps=10,
        probs_default=CP.kw(all=1),
        probs_per_depth=[CP.kw(all=0, Einsum=1), CP.kw(all=1, Scalar=(20, dict(values=[1.0], also_any=True)))],
    )
)
@hypothesis.settings(deadline=timedelta(milliseconds=400))
def test_drop_mul_ones(circuit):
    check_transform_equality(drop_mul_ones, circuit)


@given(
    circuit=st_Circuit(
        st_np.array_shapes(min_dims=0),
        must_be_explicitly_computable=True,
        max_growth_steps=3,
        probs_per_depth=[CP.kw(all=0, Rearrange=1), CP.kw(all=0, Scalar=(1, dict(values=[0.0, 1.0])))],
    )
)
def test_rearrange_of_const_to_const(circuit):
    check_transform_equality(rearrange_of_const_to_const, circuit)


@given(
    circuit=st_Circuit(
        st_np.array_shapes(min_dims=0),
        must_be_explicitly_computable=True,
        max_growth_steps=10,
        probs_default=CP.kw(all=1, Add=20),
        probs_per_depth=[CP.kw(all=0, Add=20)],
        add_must_have_weight_one=True,
    )
)
def test_flatten_adds(circuit):
    """`
    Check that if the top-most circuit is an Add, and we apply `flatten_adds`
    many times, we can get rid of the Add children.
    """
    circuit.evaluate()
    n_nodes = len(rc.all_children(circuit))  # An upper bound on the number of `flatten_adds` transforms we need

    trans = circuit
    n_iter = 0
    while isinstance(trans, rAdd) and any(isinstance(c, rAdd) for c in trans.children):
        try:
            trans = flatten_adds(trans)
        except ImpossibleRewriteError as e:
            raise RuntimeError(f"There are child adds, should not have raised {e}")
        assert trans.shape == circuit.shape
        torch.testing.assert_close(trans.evaluate(), circuit.evaluate())
        n_iter += 1
        assert n_iter <= n_nodes, "flatten_adds is not getting rid of the `Add`s"


@hypothesis.settings(
    max_examples=10, suppress_health_check=[hypothesis.HealthCheck.filter_too_much, hypothesis.HealthCheck.too_slow]
)
@given(
    circuit=st_Circuit(
        st_np.array_shapes(min_dims=0),
        must_be_explicitly_computable=True,
        max_growth_steps=3,
        probs_default=CP.kw(all=1),
        probs_per_depth=[CP.kw(all=0, Rearrange=(1, dict(do_squeeze=True)), Einsum=0)],
        rust=True,
    )
)
def test_explicit_squeeze(circuit: rc.Rearrange):
    check_transform_equality(explicit_squeeze, circuit)


@hypothesis.settings(
    max_examples=20,
    suppress_health_check=[hypothesis.HealthCheck.filter_too_much, hypothesis.HealthCheck.too_slow],
    deadline=None,
)
@given(
    circuit=st_Circuit(
        st_np.array_shapes(min_dims=0),
        must_be_explicitly_computable=True,
        max_growth_steps=3,
        probs_default=CP.kw(all=1),
        probs_per_depth=[CP.kw(all=0, Rearrange=1, Einsum=1)],
    )
)
def test_remove_noop_rearrange(circuit):
    check_transform_equality(remove_noop_rearrange, circuit)


@given(
    circuit=st_Circuit(
        st_np.array_shapes(min_dims=1),
        must_be_explicitly_computable=True,
        max_growth_steps=3,
        probs_default=CP.kw(all=1),
    ),
    keep_name=st.booleans(),
    axis=st.integers(min_value=0, max_value=5),
)
def test_remove_single_concat(circuit: Circuit, keep_name: bool, axis: int):
    assert circuit.ndim > 0
    add_circuit = Concat(circuit, axis=axis % circuit.ndim)
    trans = check_transform_equality(partial(remove_single_concat, keep_name=keep_name), add_circuit)
    if keep_name:
        assert trans.name == add_circuit.name


@st.composite
def generate_einsum_with_squeeze(draw: st.DrawFn) -> Tuple[int, int, Einsum]:
    ein = draw(
        st_Circuit(
            st_np.array_shapes(min_dims=0, max_dims=7),
            must_be_explicitly_computable=True,
            max_growth_steps=3,
            probs_default=CP.kw(all=1, Rearrange=20),
            probs_per_depth=[CP.kw(all=0, Einsum=(1, dict(min_n_children=1)))],
        )
    ).cast_einsum()
    to_unsqueeze_arg = draw(st.integers(min_value=0, max_value=len(ein.args) - 1))
    # (node, axes), other_args = ein.get_info_by(to_unsqueeze_arg)
    args = ein.args
    node, axes = args[to_unsqueeze_arg]
    to_unsqueeze_axis = draw(st.integers(min_value=0, max_value=len(axes)))

    new_node = node.unsqueeze([to_unsqueeze_axis])
    new_axes = axes[:to_unsqueeze_axis] + (ein.next_axis(),) + axes[to_unsqueeze_axis:]
    assert new_node.shape[to_unsqueeze_axis] == 1
    args[to_unsqueeze_arg] = (new_node, new_axes)

    new_ein = Einsum(*args, out_axes=ein.out_axes, name=ein.name)
    return to_unsqueeze_arg, to_unsqueeze_axis, new_ein


@hypothesis.settings(deadline=None, suppress_health_check=[hypothesis.HealthCheck.filter_too_much])
@given(
    circuit=st_Circuit(
        st_np.array_shapes(min_dims=0, max_dims=1),  # only rank 0 ever produces ident indices
        must_be_explicitly_computable=True,
        max_growth_steps=1,
        probs_default=CP.kw(all=1),
        probs_per_depth=[CP.kw(all=0, Index=1)],
    )
)
def test_remove_trivial_index(circuit: Index):
    check_transform_equality(remove_trivial_index, circuit)


@given(
    circuit=st_Circuit(
        st_np.array_shapes(min_dims=0),
        must_be_explicitly_computable=True,
        max_growth_steps=2,
        probs_default=CP.kw(all=1),
        probs_per_depth=[CP.kw(all=0, Einsum=1)],
    )
)
def test_remove_empty_einsum(circuit: Einsum):
    check_transform_equality(remove_empty_einsum, circuit)


@st.composite
def st_add_and_subset(draw: st.DrawFn) -> Tuple[Add, Set[Circuit]]:
    """
    Sample an Add circuit and a subset of its
    children (to be removed by drop_from_add,
    see test below)
    """

    circuit: Add = draw(
        st_Circuit(
            st_np.array_shapes(min_dims=0),
            must_be_explicitly_computable=True,
            max_growth_steps=1,
            probs_per_depth=[CP.kw(all=0, Add=1)],
            add_must_have_weight_one=True,
        )
    ).cast_add()
    children = circuit.children
    assert len(children) == len(set(children))

    if len(children) == 0:
        subset = []
    else:
        subset = draw(st.lists(elements=st.sampled_from(children), min_size=0, max_size=len(children), unique=True))
    return (circuit, set(subset))


@given(circuit_and_subset=st_add_and_subset())
def test_drop_from_add(circuit_and_subset: Tuple[Add, Set[Circuit]]):
    circuit: Add
    subset: Set[Circuit]
    circuit, subset = circuit_and_subset
    try:
        children = circuit.children
        hypothesis.assume(len(set(children)) == len(children))
        new_circuit = drop_from_add(circuit, subset, noop_error=True)
    except RewriteError:
        hypothesis.reject()

    assert len(subset) > 0

    assert new_circuit.shape == circuit.shape
    assert new_circuit != circuit, "The rewrite should raise ImpossibleRewriteError instead of nooping"

    if isinstance(new_circuit, Rearrange):
        new_children = new_circuit.node.children
    elif is_zero(new_circuit):
        new_children = []
    elif isinstance(new_circuit, Add):
        new_children = new_circuit.children
    else:
        assert False

    assert len(new_children) == circuit.num_children - len(
        subset
    ), f"{len(new_children)} != {circuit.num_children} - {len(subset)}; {new_circuit}"

    for deleted_circuit in subset:
        assert deleted_circuit not in new_children

    old_circuit_with_zeros = Add.from_weighted_nodes(
        *((k, (0.0 if k in subset else v)) for k, v in circuit.to_counts().items())
    )
    old_circuit_with_zeros = rc.cast_circuit(old_circuit_with_zeros, rc.TorchDeviceDtypeOp(dtype="float64"))

    new_out = new_circuit.evaluate()
    torch.testing.assert_close(old_circuit_with_zeros.evaluate().type(new_out.dtype), new_out)


# this only activates if add weight is 0, which is rare, so don't demand many examples
@hypothesis.settings(max_examples=5, suppress_health_check=[hypothesis.HealthCheck.filter_too_much])
@given(
    circuit=st_Circuit(
        st_np.array_shapes(min_dims=0),
        must_be_explicitly_computable=True,
        max_growth_steps=1,
        probs_default=CP.kw(all=1),
        probs_per_depth=[CP.kw(all=0, Add=1)],
    )
)
def test_drop_zero_add(circuit):
    check_transform_equality(drop_zero_add, circuit)


@hypothesis.example(
    circuit=Add(),
)
@given(
    circuit=st_Circuit(
        st_np.array_shapes(min_dims=0),
        must_be_explicitly_computable=True,
        max_growth_steps=1,
        probs_default=CP.kw(all=1),
        probs_per_depth=[CP.kw(all=0, Add=1).add(Add=(10, dict(min_n_children=0, max_n_children=0)))],
    )
)
def test_zero_empty_add(circuit: Add):
    check_transform_equality(zero_empty_add, circuit)


@st.composite
def st_distributed_einsum(draw: st.DrawFn):
    c: Einsum = draw(
        st_Circuit(
            st_np.array_shapes(min_dims=0),
            must_be_explicitly_computable=True,
            max_growth_steps=5,
            probs_default=CP.kw(all=1),
            probs_per_depth=[CP.kw(all=0, Einsum=(1, dict(min_n_children=2, max_n_children=2)))],
        )
    ).cast_einsum()
    assert c.num_children == 2

    ((a, a_axes), (b, b_axes)) = c.args
    out_axes = c.out_axes

    circuits = [c]

    for summand_index in range(1, 5):
        if draw(st.booleans()):
            break
        new_b = draw(
            st_Circuit(
                b.shape,
                must_be_explicitly_computable=True,
                max_growth_steps=5,
            )
        )

        c_p = Einsum((a, a_axes), (new_b, b_axes), out_axes=out_axes)
        assert out_axes == c_p.out_axes
        circuits.append(c_p)

    return Add(*circuits)


@given(einsum_circuit=st_distributed_einsum(), is_lhs=st.booleans())
def test_factor_add_of_mul_to_mul_of_add(einsum_circuit: Add, is_lhs: bool):
    check_transform_equality(partial(factor_add_of_mul_to_mul_of_add, is_lhs=is_lhs), einsum_circuit)


@st.composite
def st_concat_add(draw: st.DrawFn):
    circuit = draw(
        st_Circuit(
            st_np.array_shapes(min_dims=1),
            must_be_explicitly_computable=True,
            max_growth_steps=5,
            probs_per_depth=[
                CP.kw(all=0, Add=1),
            ],
        )
    )

    assert isinstance(circuit, Add)
    childrens_shapes = [child.shape for child in circuit.children]
    adds = [circuit]

    concat_axis = draw(st.integers(min_value=0, max_value=len(circuit.shape) - 1))

    concat_length = draw(st.integers(min_value=1, max_value=5))

    for add_idx in range(1, concat_length):
        children: List[Circuit] = []

        for child_shape in childrens_shapes:
            children.append(
                draw(
                    st_Circuit(
                        child_shape,
                        must_be_explicitly_computable=True,
                        max_growth_steps=5,
                    )
                )
            )

        adds.append(Add.from_weighted_nodes(*((child, draw(st_c.bounded_floats)) for child in children)))

    return Concat(*adds, axis=concat_axis)


# @hypothesis.settings(suppress_health_check=[hypothesis.HealthCheck.filter_too_much, hypothesis.HealthCheck.too_slow])
# @given(circuit=st_concat_add())
# def test_distribute_concat(circuit):
#     summand_to_pull_out = 0
#     check_transform_equality(lambda c: distribute_concat(c, summand_to_pull_out), circuit)


@st.composite
def st_add_concat(draw: st.DrawFn):
    circuit = draw(
        st_Circuit(
            st_np.array_shapes(min_dims=1),
            must_be_explicitly_computable=True,
            max_growth_steps=20,
            probs_per_depth=[
                CP.kw(all=0, Concat=1),
            ],
            add_must_have_weight_one=True,
        )
    )

    assert isinstance(circuit, Concat)
    childrens_shapes = [child.shape for child in circuit.children]
    concats = [circuit]

    sum_length = draw(st.integers(min_value=0, max_value=5))
    if sum_length == 0:
        concats = []

    for concat_idx in range(1, sum_length):
        children: List[Circuit] = []

        for child_shape in childrens_shapes:
            children.append(
                draw(
                    st_Circuit(
                        child_shape,
                        must_be_explicitly_computable=True,
                        max_growth_steps=5,
                    )
                )
            )

        concats.append(
            Concat(
                *children,
                # tuple((child, draw(st_c.bounded_floats)) for child in children),
                axis=circuit.axis,
            )
        )

    return Add.from_weighted_nodes(*[(concat, draw(st_c.bounded_floats)) for concat in concats])


@hypothesis.settings(
    suppress_health_check=[hypothesis.HealthCheck.filter_too_much, hypothesis.HealthCheck.too_slow], deadline=None
)
@given(circuit=st_add_concat())
def test_pull_through_add_concat(circuit):
    check_transform_equality(pull_through_add_concat, circuit)


@given(
    circuit=st_Circuit(
        st_np.array_shapes(min_dims=0),
        must_be_explicitly_computable=True,
        max_growth_steps=10,
        growth_order="breadth_first",
        probs_default=CP.kw(all=1),
        probs_per_depth=[CP.kw(all=0, Add=(1, dict(max_n_children=3))), CP.kw(all=0, Index=1)],
        add_must_have_weight_one=True,
    )
)
@example(circuit=Add(Index(node=Scalar(0.0, (1,)), index=(0,)), Index(node=Scalar(1.0, (1,)), index=(0,))))
@example(
    circuit=Add(
        Index(node=Scalar(0.0, shape=(25, 30)), index=(slice(2, 12), 3)),
        Index(node=Scalar(1.0, shape=(25, 30)), index=(slice(2, 12), 3)),
    )
)
def test_pull_through_add_index(circuit: Add):
    assert all(isinstance(c, Index) for c in circuit.children)
    check_transform_equality(pull_through_add_index, circuit)


@hypothesis.settings(suppress_health_check=[hypothesis.HealthCheck.filter_too_much])
@given(
    circuit=st_Circuit(
        st_np.array_shapes(min_dims=0),
        must_be_explicitly_computable=True,
        max_growth_steps=3,
        probs_default=CP.kw(all=1),
        probs_per_depth=[
            CP.kw(all=0, Add=1),
            CP.kw(all=1, Rearrange=(50, dict(do_squeeze=False))),
        ],
        add_must_have_weight_one=True,
    )
)
def test_pull_through_add_rearrange(circuit):
    check_transform_equality(pull_through_add_rearrange, circuit)


@given(
    circuit=st_Circuit(
        st_np.array_shapes(min_dims=0),
        must_be_explicitly_computable=True,
        max_growth_steps=3,
        probs_per_depth=[
            CP.kw(all=0, Einsum=1),
            CP.kw(all=1, Rearrange=(50, dict(do_split=False, do_unsqueeze=False, do_repeat=False, do_squeeze=False))),
        ],
    )
)
@hypothesis.example(
    circuit=Einsum(
        (
            rc.Rearrange.from_string(Scalar(0.0, shape=(1, 2, 1), name=""), "a:1 c:2 b:1 -> a b c"),
            (0, 0, 1),
        ),
        (Scalar(0.0, shape=(), name=""), ()),
        (Scalar(0.0, shape=(1,), name=""), (0,)),
        out_axes=(0, 1),
        name="given_test",
    ),
)
def test_fuse_einsum_permute(circuit: Einsum):
    check_transform_equality(fuse_einsum_permute, circuit)


@given(
    circuit=st_Circuit(
        st_np.array_shapes(min_dims=0),
        must_be_explicitly_computable=True,
        max_growth_steps=3,
        probs_per_depth=[
            CP.kw(all=0, Einsum=1),
            CP.kw(all=1, Rearrange=(50, dict(do_split=False, do_unsqueeze=True, do_repeat=True))),
        ],
    )
)
def test_fuse_einsum_rearrange(circuit: Einsum):
    check_transform_equality(fuse_einsum_rearrange, circuit)


@given(
    circuit=st_Circuit(
        st_np.array_shapes(min_dims=0),
        must_be_explicitly_computable=True,
        max_growth_steps=5,
        probs_default=CP.kw(all=1),
        probs_per_depth=[CP.kw(all=0, Rearrange=1), CP.kw(all=1, Einsum=50)],
    )
)
@hypothesis.example(
    circuit=rc.Rearrange.from_string(
        Einsum(
            (Scalar(0.0, shape=(2, 1), name=""), (0, 1)),
            out_axes=(0, 1),
        ),
        "b:2 a:1 -> a b",
    ),
)
@mark_not_interesting_if(ImpossibleRewriteError)
def test_fuse_permute_einsum(circuit):
    check_transform_equality(fuse_permute_einsum, circuit)


@given(
    circuit=st_Circuit(
        st_np.array_shapes(min_dims=0),
        must_be_explicitly_computable=True,
        max_growth_steps=5,
        probs_default=CP.kw(all=1),
        probs_per_depth=[
            CP.kw(all=0, Rearrange=1),
            CP.kw(all=0, Rearrange=1),
        ],
    )
)
@mark_not_interesting_if(ImpossibleRewriteError)
def test_fuse_rearrange(circuit):
    check_transform_equality(fuse_rearrange, circuit)


@given(
    circuit=st_Circuit(
        st_np.array_shapes(min_dims=0, max_dims=3),
        must_be_explicitly_computable=True,
        max_growth_steps=3,
        probs_default=CP.kw(all=1),
        probs_per_depth=[CP.kw(all=0, Einsum=(1, dict(min_n_children=1, max_n_children=1))), CP.kw(all=1, Einsum=50)],
    )
)
@mark_not_interesting_if(ImpossibleRewriteError)
def test_fuse_single_einsum(circuit):
    check_transform_equality(fuse_single_einsum, circuit)


@hypothesis.settings(deadline=timedelta(milliseconds=300))
@given(
    circuit=st_Circuit(
        st_np.array_shapes(min_dims=0),
        must_be_explicitly_computable=True,
        max_growth_steps=5,
        growth_order="breadth_first",
        probs_default=CP.kw(all=1),
        probs_per_depth=[
            CP.kw(all=0, Einsum=1),
            CP.kw(all=0, Einsum=1).add(Einsum=(50, dict(min_n_children=1, max_n_children=1))),
            CP.kw(all=1, Einsum=50),
        ],
    )
)
@mark_not_interesting_if(ImpossibleRewriteError)
def test_fuse_einsum_single(circuit: Einsum):
    check_transform_equality(fuse_einsum_single, circuit)


def test_fuse_einsum_single_simple():
    c = rEinsum(
        (rEinsum((rScalar(1.0), ()), out_axes=()), ()),
        (rEinsum((rScalar(1.0, (1,)), (0,)), out_axes=(0,)), (0,)),
        (rEinsum((rScalar(1.0, (1,)), (0,)), (rScalar(0.5, (1,)), (0,)), out_axes=(0,)), (0,)),
        (rEinsum((rScalar(1.0, (1,)), (0,)), (rScalar(0.5, (1,)), (1,)), out_axes=(0, 1)), (0, 1)),
        (rEinsum((rScalar(1.0, (1,)), (0,)), out_axes=(0,)), (1,)),
        out_axes=(1,),
    )
    r = rEinsum(
        (rScalar(1.0), ()),
        (rScalar(1.0, (1,)), (0,)),
        (rEinsum((rScalar(1.0, (1,)), (0,)), (rScalar(0.5, (1,)), (0,)), out_axes=(0,)), (0,)),
        (rEinsum((rScalar(1.0, (1,)), (0,)), (rScalar(0.5, (1,)), (1,)), out_axes=(0, 1)), (0, 1)),
        (rScalar(1.0, (1,)), (1,)),
        out_axes=(1,),
    )
    assert r == fuse_einsum_single(c)


def test_split_einsum_concat():
    a = Array(torch.randn(3, 4, 5), name="a")
    b = Array(torch.randn(3, 5, 7), name="b")
    c = Array(torch.randn(3, 4, 2), name="c")
    d = Array(torch.randn(3, 1, 5), name="d")
    e = Array(torch.randn(3, 2, 5), name="e")
    f = Array(torch.randn(3, 3, 5), name="f")

    # evaluator = MemoizedFn(evaluate_fn(dtype=torch.float64))
    for out in ["x w v y z", "x w v y", "z", "z w", "w", "y w", "y z", "y x"]:
        for concat in [Concat(a, c, axis=-1), Concat(a, f, d, e, axis=1)]:
            v = Einsum.from_einsum_string(f"x y z, x w v -> {out}", concat, b)

            all_groups: Tuple[List[List[int]], ...] = ([[0]], [[1]], [[0], [1]], [[1], [0]])
            if len(concat.children) > 2:
                all_groups += ([[1, 2]], [[3], [1, 2]], [[0], [2]], [[0], [3]])
            for groups in all_groups:
                torch.testing.assert_close(
                    v.evaluate(), split_einsum_concat(v, element_index=0, index_groups=groups).evaluate()
                )

    for out in ["x w y z", "x w y", "z", "z w", "w", "y w", "y z", "y x"]:
        for concat in [Concat(a, c, axis=-1), Concat(a, f, d, e, axis=1)]:
            l = Array(torch.randn(6, *concat.shape[1:]), name="z")
            v = Einsum.from_einsum_string(f"x y z, w y z -> {out}", concat, l)

            all_groups = ([[0]], [[1]], [[0], [1]], [[1], [0]])
            if len(concat.children) > 2:
                all_groups += ([[1, 2]], [[3], [0, 1, 2]], [[0], [2]], [[0], [3]])
            for groups in all_groups:
                torch.testing.assert_close(
                    v.evaluate(), split_einsum_concat(v, element_index=0, index_groups=groups).evaluate()
                )


@hypothesis.settings(
    suppress_health_check=[hypothesis.HealthCheck.filter_too_much, hypothesis.HealthCheck.too_slow],
    deadline=timedelta(seconds=6),
)
@given(
    circuit=st_Circuit(
        st_np.array_shapes(min_dims=0),
        must_be_explicitly_computable=True,
        max_growth_steps=4,
        growth_order="breadth_first",
        probs_default=CP.kw(all=1),
        probs_per_depth=[CP.kw(all=0, Einsum=(1, dict(min_n_children=1))), CP.kw(all=0, Add=1)],
    ),
    element_index=st.integers(0, 8),
)
@hypothesis.example(
    circuit=Einsum(
        (Add(), ()),
        (Add.from_weighted_nodes((Scalar(0.0, (1,)), 0.0)), (0,)),
        out_axes=(0,),
    ),
    element_index=0,
)
def test_distribute(circuit: Einsum, element_index: int):
    assert len(circuit.args) > 0
    check_transform_equality(partial(distribute_old, element_index=element_index), circuit)


@given(
    circuit=st_Circuit(
        st_np.array_shapes(min_dims=0),
        must_be_explicitly_computable=True,
        max_growth_steps=4,
        probs_default=CP.kw(all=1),
        probs_per_depth=[
            CP.kw(all=0, Rearrange=(1, dict(do_split=False, do_unsqueeze=False, do_repeat=False, do_squeeze=False)))
        ],
    )
)
def test_permute_to_einsum(circuit: Rearrange):
    assert circuit.spec.is_permute()
    check_transform_equality(permute_to_einsum, circuit)


@given(
    circuit=st_Circuit(
        st_np.array_shapes(min_dims=0),
        must_be_explicitly_computable=True,
        max_growth_steps=4,
        probs_default=CP.kw(all=1),
        probs_per_depth=[
            CP.kw(all=0, Rearrange=(1, dict(do_split=False, do_unsqueeze=False, do_repeat=False, do_squeeze=False))),
            CP.kw(all=0, Add=1),
        ],
    )
)
@mark_not_interesting_if(NotImplementedError)
def test_push_down_permute_via_einsum(circuit: Rearrange):
    assert circuit.spec.is_permute()
    assert isinstance(circuit.node, Add)
    check_transform_equality(push_down_permute_via_einsum, circuit)


ExplicitReduceArgs = Tuple[Einsum, int, int, Sequence[TorchAxisIndex]]


@st.composite
def st_explicit_reduce_args(draw: st.DrawFn, must_be_explicitly_computable: bool = True) -> ExplicitReduceArgs:
    st_circuit = st_Circuit(
        st_np.array_shapes(min_dims=0),
        must_be_explicitly_computable=must_be_explicitly_computable,
        max_growth_steps=4,
        probs_default=CP.kw(all=1),
        probs_per_depth=[CP.kw(all=0, Einsum=(1, dict(min_n_children=1)))],
    )
    circuit = draw(st_circuit)
    assert isinstance(circuit, Einsum) and len(circuit.args) > 0

    which = draw(st.integers(min_value=0, max_value=len(circuit.args) - 1))
    circ = circuit.children[which]
    if circ.ndim == 0:
        # Attempt the rewrite anyways, should raise ImpossibleRewriteError
        return circuit, 0, which, []

    axis = draw(st.integers(min_value=0, max_value=circ.ndim - 1))

    idxs: List[TorchAxisIndex] = []
    end = 0
    while end != circ.shape[axis]:
        if draw(st.booleans()):
            new_end = draw(st.integers(min_value=end, max_value=circ.shape[axis]))
            idxs.append(slice(end, new_end))
            end = new_end
        else:
            idxs.append(end)
            end += 1

    return circuit, axis, which, idxs


@given(
    circuit=st_Circuit(
        st.one_of(*[st.just(x) for x in [(2, 2), (3, 3, 3), (2, 2, 3, 3), (3, 3, 3, 2)]]),  # upsample diags!
        must_be_explicitly_computable=True,
        max_growth_steps=2,
        probs_default=CP.kw(all=1),
        probs_per_depth=[CP.kw(all=0, Einsum=(1, dict(min_n_children=1)))],
    ),
)
def test_extract_output_diag(circuit: Einsum):
    check_transform_equality(extract_output_diags, circuit)


@st.composite
def st_einsum_with_diag_index(draw: st.DrawFn, must_be_explicitly_computable: bool = True) -> tuple[Einsum, int]:
    circuit = draw(
        st_Circuit(
            st_np.array_shapes(min_dims=0),
            must_be_explicitly_computable=must_be_explicitly_computable,
            max_growth_steps=2,
            probs_default=CP.kw(all=1),
            probs_per_depth=[CP.kw(all=0, Einsum=(1, dict(min_n_children=1)))],
        )
    )
    assert isinstance(circuit, Einsum)

    index_options = [i for i, axes in enumerate(circuit.all_input_axes()) if len(set(axes)) < len(axes)]
    hypothesis.assume(len(index_options) > 0, "we need some duplicate axes")
    index = index_options[draw(st.integers(min_value=0, max_value=len(index_options) - 1))]

    return circuit, index


@given(
    einsum_diag=st_einsum_with_diag_index(must_be_explicitly_computable=True),
)
@hypothesis.settings(suppress_health_check=[hypothesis.HealthCheck.filter_too_much, hypothesis.HealthCheck.too_slow])
def test_extract_input_diag(einsum_diag: Tuple[Einsum, int]):
    einsum, diag_index = einsum_diag

    check_transform_equality(lambda x: extract_input_diags(x, diag_index), einsum)


@given(explicit_reduce_args=st_explicit_reduce_args())
@hypothesis.settings(deadline=None)
@mark_not_interesting_if(NotImplementedError)
def test_explicit_reduce(explicit_reduce_args: ExplicitReduceArgs):
    check_transform_equality(lambda c: explicit_reduce(c, *explicit_reduce_args[1:]), explicit_reduce_args[0])


@given(
    circuit=st_Circuit(
        st_np.array_shapes(min_dims=0),
        must_be_explicitly_computable=True,
        max_growth_steps=4,
        probs_default=CP.kw(all=1),
        probs_per_depth=[CP.kw(all=0, Einsum=1), CP.kw(all=1, Einsum=70)],
    )
)
@example(circuit=Einsum((Einsum(out_axes=()), ()), out_axes=()))
@mark_not_interesting_if(
    ValueError, message="einsum(): subscript in subscript list is not within the valid range [0, 52)"
)
def test_einsum_flatten_bans_noop(circuit: Einsum):
    check_transform_equality(lambda x: einsum_flatten_bans_noop(x), circuit)


ExtractAddArgs = Tuple[Add, Add]


@st.composite
def st_extract_add_args(draw: st.DrawFn, must_be_explicitly_computable: bool = True) -> ExtractAddArgs:
    st_circuit = st_Circuit(
        st_np.array_shapes(min_dims=1),
        must_be_explicitly_computable=must_be_explicitly_computable,
        max_growth_steps=2,
        probs_per_depth=[CP.kw(all=0, Add=1)],
    )
    circuit = draw(st_circuit)
    assert isinstance(circuit, Add)

    all_sub = dedup_with_order(
        draw(
            st.lists(
                elements=st.integers(min_value=0, max_value=len(circuit.children) - 1),
                max_size=len(circuit.children) - 1,
            )
        )
    )

    ch = circuit.children

    return circuit, Add(*(ch[i] for i in all_sub))


@given(extract_add_args=st_extract_add_args())
def test_extract_add(extract_add_args: ExtractAddArgs):
    circuit, sub = extract_add_args

    check_transform_equality(lambda x: extract_add(x, sub), circuit)


@hypothesis.settings(deadline=None)
@given(extract_add_args=st_extract_add_args())
def test_extract_add_by_match(extract_add_args: ExtractAddArgs):
    circuit, sub = extract_add_args
    matcher = Matcher(set(sub.children))

    check_transform_equality(lambda x: extract_add_by_match(x, matcher), circuit)


NestedEinsumAxesPermuteArgs = Tuple[Einsum, int, List[int]]


@st.composite
def st_nested_einsum_axes_permute_args(
    draw: st.DrawFn, must_be_explicitly_computable: bool = True
) -> NestedEinsumAxesPermuteArgs:
    st_circuit = st_Circuit(
        st_np.array_shapes(min_dims=0),
        must_be_explicitly_computable=must_be_explicitly_computable,
        max_growth_steps=2,
        probs_default=CP.kw(all=1),
        probs_per_depth=[CP.kw(all=0, Einsum=(1, dict(min_n_children=2, max_n_children=6))), CP.kw(all=1, Einsum=40)],
    )
    circuit = draw(st_circuit)
    assert isinstance(circuit, Einsum)

    einsums = [(i, x) for i, x in enumerate(circuit.children) if isinstance(x, Einsum)]
    hypothesis.assume(len(einsums) > 0, "no einsums")
    which = draw(st.integers(min_value=0, max_value=len(einsums) - 1))
    idx, einsum = einsums[which]
    perm = draw(st.permutations(range(einsum.ndim)))

    return circuit, idx, perm


@given(nested_einsum_axes_permute_args=st_nested_einsum_axes_permute_args())
@hypothesis.settings(suppress_health_check=[hypothesis.HealthCheck.filter_too_much, hypothesis.HealthCheck.too_slow])
def test_nested_einsum_axes_permute(nested_einsum_axes_permute_args: NestedEinsumAxesPermuteArgs):
    circuit, element_index, permutation = nested_einsum_axes_permute_args

    def wrap_test(x: rEinsum):
        out = nested_einsum_axes_permute(x, element_index=element_index, permutation=permutation)
        if out == x:
            raise ImpossibleRewriteError()
        return out

    check_transform_equality(wrap_test, circuit)


def test_nested_einsum_permute_dups_to_eq_manual():
    x = Tag.new_with_random_uuid(Scalar(0.0, (2, 3)))
    y = Tag.new_with_random_uuid(Scalar(0.0, (4, 5, 3)))
    Tag.new_with_random_uuid(Scalar(0.0, (6, 4, 5)))

    x_y = Einsum((x, (0, 1)), (y, (2, 3, 1)), out_axes=(2, 1, 3, 0))
    x_y_other = Einsum((x, (0, 1)), (y, (2, 3, 1)), out_axes=(0, 1, 3, 2))

    mul = Einsum((x_y, (2, 1, 3, 0)), (x_y_other, (7, 5, 6, 2)), out_axes=(3, 2, 6, 7))

    assert len(set(mul.children)) == 2
    assert len(set(nested_einsum_permute_dups_to_eq(mul).children)) == 1


MultiplyAxesByIdentityArgs = Tuple[Circuit, List[int], List[int]]


@st.composite
def st_multiply_axes_by_identity_args(
    draw: st.DrawFn, must_be_explicitly_computable: bool = True
) -> MultiplyAxesByIdentityArgs:
    st_circuit = st_Circuit(
        st_np.array_shapes(min_dims=1),
        must_be_explicitly_computable=must_be_explicitly_computable,
        max_growth_steps=2,
        probs_per_depth=[CP.kw(all=0, leaves=1)],
    )
    circuit = draw(st_circuit)

    all_axes = dedup_with_order(
        draw(st.lists(elements=st.integers(min_value=0, max_value=circuit.ndim - 1), max_size=circuit.ndim - 1))
    )
    batch_axes = [
        i
        for i in dedup_with_order(
            draw(st.lists(elements=st.integers(min_value=0, max_value=circuit.ndim - 1), max_size=circuit.ndim - 1))
        )
        if i not in all_axes
    ]

    return circuit, all_axes, batch_axes


@given(multiply_axes_by_identity_args=st_multiply_axes_by_identity_args())
def test_multiply_axes_by_identity(multiply_axes_by_identity_args: MultiplyAxesByIdentityArgs):
    circuit, axes, batch_axes = multiply_axes_by_identity_args

    check_transform_equality(lambda x: multiply_axes_by_identity(x, axes, batch_axes=batch_axes), circuit)


SplitToConcatForBatchArgs = Tuple[Circuit, int, int]


@st.composite
def st_split_to_concat_for_batch_args(
    draw: st.DrawFn, must_be_explicitly_computable: bool = True
) -> SplitToConcatForBatchArgs:
    st_circuit = st_Circuit(
        st_np.array_shapes(min_dims=1),
        must_be_explicitly_computable=must_be_explicitly_computable,
        probs_per_depth=[CP.kw(all=0, leaves=1)],
    )
    circuit = draw(st_circuit)

    axis = draw(st.integers(0, circuit.ndim - 1))
    extra_mul = 5 if draw(st.booleans()) else 1
    batch_size = draw(st.integers(1, circuit.shape[axis] * extra_mul))
    return circuit, batch_size, axis


@given(split_to_concat_for_batch_args=st_split_to_concat_for_batch_args())
def test_split_to_concat_for_batch(split_to_concat_for_batch_args: SplitToConcatForBatchArgs):
    circuit, batch_size, axis = split_to_concat_for_batch_args
    check_transform_equality(lambda x: split_to_concat_for_batch(x, batch_size=batch_size, axis=axis), circuit)


def test_apply_in_traversal():
    circ = P(
        """
0 Einsum abc->bca
  1 Add
    2 Add
      3 [2,2,2] Scalar 1.0
    """
    )
    trav = new_traversal(start_depth=0, end_depth=3)
    result = apply_in_traversal(circ, trav, simp)
    result.print()
    assert result == P(
        """
0 Rearrange a b c -> b c a
  1 Add
    2 [2, 2, 2] Scalar 1"""
    )
    replace_outside_traversal_symbols(circ, trav)[0].print()
    replace_outside_traversal_symbols(circ, trav, lambda x: "nameyname")[0].print()
    replace_outside_traversal_symbols(circ, trav, lambda x: None)[0].print()



import pytest
import torch

from interp.tools.indexer import I
from rust_circuit import Array, MiscInputItemOutOfBoundsError


def test_convenience_methods():
    a = torch.tensor([[0.0, 1], [2, 3]])
    b = torch.tensor([[0.0, 10], [20, 30]])
    circ = Array(a)
    circ2 = Array(b)

    # circ.add(circ2).print(True, True)
    torch.testing.assert_close(circ.add(circ2).evaluate(), a + b)
    # circ.sub(circ2).print(True, True)
    torch.testing.assert_close(circ.sub(circ2).evaluate(), a - b)
    # circ.mean(0).print(True, True)
    # circ.mean().print(True, True)
    # circ.mean(1, "hooo").print(True, True)
    # circ.min(1, "hooo").print(True, True)
    # circ.max(1, "hooo").print(True, True)

    torch.testing.assert_close(circ.mean().evaluate(), a.mean())
    torch.testing.assert_close(circ.sum().evaluate(), a.sum())
    torch.testing.assert_close(circ.min().evaluate(), a.amin())
    torch.testing.assert_close(circ.max().evaluate(), a.amax())
    torch.testing.assert_close(circ.max(axis=0).evaluate(), a.amax(dim=0))
    torch.testing.assert_close(circ.max(axis=1).evaluate(), a.amax(dim=1))
    torch.testing.assert_close(circ.max(axis=-1).evaluate(), a.amax(dim=-1))
    torch.testing.assert_close(circ.max(axis=-2).evaluate(), a.amax(dim=-2))
    torch.testing.assert_close(circ.max(axis=(0, 1)).evaluate(), a.amax(dim=(0, 1)))
    torch.testing.assert_close(circ.max(axis=(0, -1)).evaluate(), a.amax(dim=(0, -1)))
    torch.testing.assert_close(circ.max(axis=(-2, -1)).evaluate(), a.amax(dim=(-2, -1)))
    torch.testing.assert_close(circ.mean(axis=(0, 1)).evaluate(), a.mean(dim=(0, 1)))
    torch.testing.assert_close(circ.mean(axis=0).evaluate(), a.mean(dim=0))
    torch.testing.assert_close(circ.mean(axis=1).evaluate(), a.mean(dim=1))
    torch.testing.assert_close(circ.mean(axis=-1).evaluate(), a.mean(dim=-1))

    torch.testing.assert_close(circ.mul(circ2).evaluate(), a * b)
    torch.testing.assert_close(circ.index(I[0:1, 1]).evaluate(), a[0:1, 1])


def test_squeeze():
    circ = Array.randn(1, 3, 1)
    circ.squeeze(0).print()
    circ.squeeze([0, 2]).print()
    with pytest.raises(RuntimeError):
        circ.squeeze([0, 1]).print()
    with pytest.raises(MiscInputItemOutOfBoundsError):
        circ.squeeze(3).print()


if __name__ == "__main__":
    test_convenience_methods()



from typing import Tuple

import hypothesis
import hypothesis.extra.numpy as st_np
import hypothesis.strategies as st
import torch
from torch.testing import assert_close

import rust_circuit as rc
from interp.circuit.algebric_rewrite import (
    MulRearrangeSpec,
    MulRearrangeSpecSub,
    MulRestType,
    NamedItem,
    get_einsum_ordering_items,
)
from interp.circuit.circuit_utils import cast_circuit
from interp.circuit.test_algebric_rewrite import partitioned_or_rest
from interp.circuit.testing.topdown_circuit import CircuitProbs as CP
from interp.circuit.testing.utils import mark_not_interesting_if
from rust_circuit import Add, Einsum, NestAddsSpecSub, NestEinsumsSpecSub
from rust_circuit import NestRest as Rest
from rust_circuit import add_flatten, nest_adds, nest_einsums
from rust_circuit.interop_rust import py_to_rust

from .util import st_Circuit

NestEinsumArgs = Tuple[Einsum, NestEinsumsSpecSub]


@st.composite
def st_nest_einsum_args(draw: st.DrawFn, must_be_explicitly_computable: bool = True) -> NestEinsumArgs:
    """TODO: currently does not generate any :class:`NamedItem` s or terminating node specs."""
    st_circuit = st_Circuit(
        st_np.array_shapes(min_dims=0),
        must_be_explicitly_computable=must_be_explicitly_computable,
        max_growth_steps=4,
        probs_default=CP.kw(all=1, Einsum=70),
        probs_per_depth=[CP.kw(all=0, Einsum=1), CP.kw(all=1, Einsum=70)],
        possibly_different_name=True,
    )
    circuit = draw(st_circuit).cast_einsum()

    # TODO: improve this testing as needed!
    spec: MulRearrangeSpecSub
    if (n := len(rc.einsum_flatten(circuit).args)) == 0:
        spec = ()
    else:
        idxs = draw(st.permutations(range(n)))
        spec, _ = draw(partitioned_or_rest(idxs, can_have_rest=True))

    def convert(spec: MulRearrangeSpec) -> NestEinsumsSpecSub:
        if isinstance(spec, int):
            return spec
        if isinstance(spec, tuple):
            return [convert(s) for s in spec]
        if isinstance(spec, MulRestType):
            return Rest()
        assert not isinstance(spec, NamedItem)

        assert False

    spec_out = convert(spec)

    circuit = rc.cast_circuit(circuit, rc.TorchDeviceDtypeOp(dtype="float64"))

    return circuit.cast_einsum(), spec_out


NestAddArgs = Tuple[Add, NestAddsSpecSub]


# TODO: dedup with above
@st.composite
def st_nest_add_args(draw: st.DrawFn, must_be_explicitly_computable: bool = True) -> NestAddArgs:
    """TODO: currently does not generate any :class:`NamedItem` s or terminating node specs."""
    st_circuit = st_Circuit(
        st_np.array_shapes(min_dims=0),
        must_be_explicitly_computable=must_be_explicitly_computable,
        max_growth_steps=4,
        probs_default=CP.kw(all=1, Add=(70, dict(use_weights=False))),
        probs_per_depth=[
            CP.kw(all=0, Add=(1, dict(use_weights=False))),
            CP.kw(all=1, Add=(70, dict(use_weights=False))),
        ],
        possibly_different_name=True,
    )
    circ = draw(st_circuit)

    circ = rc.cast_circuit(circ, rc.TorchDeviceDtypeOp(dtype="float64")).cast_add()

    # TODO: improve this testing as needed!
    spec: MulRearrangeSpecSub
    if (n := len(add_flatten(circ).children)) == 0:
        spec = ()
    else:
        idxs = draw(st.permutations(range(n)))
        spec, _ = draw(partitioned_or_rest(idxs, can_have_rest=True))

    def convert(spec: MulRearrangeSpec) -> NestAddsSpecSub:
        if isinstance(spec, int):
            return spec
        if isinstance(spec, tuple):
            return [convert(s) for s in spec]
        if isinstance(spec, MulRestType):
            return Rest()
        assert not isinstance(spec, NamedItem)

        assert False

    spec_out = convert(spec)

    return circ, spec_out


@hypothesis.given(nest_einsum_args=st_nest_einsum_args())
@mark_not_interesting_if(
    ValueError, message="einsum(): subscript in subscript list is not within the valid range [0, 52)"
)
def test_nest_einsums_hypothesis(nest_einsum_args: NestEinsumArgs):
    circuit, spec = nest_einsum_args
    nested = nest_einsums(circuit, spec)
    # circuit.print()
    # nested.print()
    assert_close(
        nested.evaluate(),
        circuit.evaluate(),
        rtol=1e-4,
        atol=1e-4,
    )


@hypothesis.given(nest_add_args=st_nest_add_args())
def test_nest_adds_hypothesis(nest_add_args: NestAddArgs):
    circuit, spec = nest_add_args
    nested = nest_adds(circuit, spec)
    # circuit.print()
    # nested.print()
    assert_close(
        nested.evaluate(),
        circuit.evaluate(),
        rtol=1e-4,
        atol=1e-4,
    )



# %%

from typing import List

import pytest

import rust_circuit as rc
import rust_circuit.module_library as mod_l
from rust_circuit.model_rewrites import (
    To,
    clear_block_module,
    configure_transformer,
    flatten_res,
    pull_up_bias,
    pull_up_head_split,
    push_down_transformer_weights,
    split_by_head,
    strip_arr,
)
from rust_circuit.py_utils import S

# %%

# we'll start by getting a model
# we'll use biases for demoing for now
params = mod_l.TransformerParams(mod_l.TransformerBlockParams(attn_bias=True, mlp_output_bias=True), num_layers=2)
c, _, _ = params.garbage_call(num_heads=2)  # 2 heads to avoid too much clutter

# note that we bind here - all of the below config funcs are intended to be
# applied to this t.bind_w circuit.
# You might want to use `c.update("t.bind_w", ...)` in various cases
c = c.get_unique("t.bind_w")

# %%

# and some printers
basic_term_early: List[rc.MatcherIn] = ["a", "m", "ln", "a.head.on_inp"]
printer = rc.PrintOptions(
    bijection=False,
    traversal=rc.new_traversal(term_early_at=rc.Matcher(*basic_term_early)),
    colorer=rc.PrintOptions.type_colorer(),
)
less_term_early: List[rc.MatcherIn] = [
    *basic_term_early,
    rc.Matcher.regex(r"^a\d+((.h\d+)|(.norm))$"),
    rc.Matcher.regex(r"^m\d+(.norm)?(.p_bias)?$"),
    "final.norm",
]
less_printer = printer.evolve(
    traversal=rc.new_traversal(term_early_at=rc.Matcher(*less_term_early, rc.Matcher.regex(r"^a\d+(.p_bias)?$")))
)
less_printer_split = printer.evolve(traversal=rc.new_traversal(term_early_at=rc.Matcher(*less_term_early)))

# %%

# note that the model will be loaded with the weights all bound in one big
# module - this is nice for loading and some other stuff, but not typically
# nice for interp
printer.print(c)

# %%

# now we can configure our model in a variety of ways!

# here are the default args
# see docs for more details
out = configure_transformer(
    c,
    to=To.ATTN_MLP_NORM,  # push down weights to bind at attn/mlp/norm
    use_pull_up_bias=True,
    use_strip_arr=True,
    split_by_head_config=None,
    use_pull_up_head_split=False,
    use_clear_block_module=True,
    use_flatten_res=False,
    flatten_components=False,
    check_valid=True,
)
printer.print(out)

# %%

# quite pretty IMO
less_printer.print(out)

# %%

# now some examples/tests

# %%


no_pull = configure_transformer(c, use_pull_up_bias=False)
printer.print(no_pull)

# %%

# side by side
less_printer.print(out)
less_printer.print(no_pull)  # biases are inside

# %%

# be warned, some args require specific settings of other args!
with pytest.raises(AssertionError) as exc0:
    configure_transformer(c, split_by_head_config="full")
print(exc0.exconly())

# %%

split = configure_transformer(c, to=To.ATTN_HEAD_MLP_NORM, split_by_head_config="full")
printer.print(split)

# %%

less_printer_split.print(split)

# %%

# this moves the norm to be per head - this can be nice if you want to think of
# each head + norm as a separate function
split_pull = configure_transformer(
    c, to=To.ATTN_HEAD_MLP_NORM, split_by_head_config="full", use_pull_up_head_split=True
)
printer.print(split_pull)

# %%

less_printer_split.print(split)
less_printer_split.print(split_pull)

# %%

# I don't typically recommend using flat, but you can if you want
flat = configure_transformer(c, use_flatten_res=True)
less_printer.print(flat)

# %%

# if we pull up biases, by default they still aren't flattened in
# (same as above, use_pull_up_bias is default)
flat = configure_transformer(c, use_flatten_res=True, use_pull_up_bias=True)
# printer.print(flat)
less_printer.print(flat)

# %%

# but we can do that by using flatten_components
flat = configure_transformer(c, use_flatten_res=True, use_pull_up_bias=True, flatten_components=True)
# printer.print(flat)
less_printer.print(flat)

# %%

# attn heads also aren't flattened in even if pulled up
split_pull_some_flat = configure_transformer(
    c, to=To.ATTN_HEAD_MLP_NORM, split_by_head_config="full", use_pull_up_head_split=True, use_flatten_res=True
)
# printer.print(split_pull_some_flat)
less_printer_split.print(split_pull_some_flat)

# %%

# but we can flatten_components
split_pull_flat = configure_transformer(
    c,
    to=To.ATTN_HEAD_MLP_NORM,
    split_by_head_config="full",
    use_pull_up_head_split=True,
    use_flatten_res=True,
    flatten_components=True,
)
# printer.print(split_pull_flat)
less_printer_split.print(split_pull_flat)

# %%

# the head splitting also supports some more detailed splitting.
split_first = configure_transformer(
    c,
    to=To.ATTN_HEAD_MLP_NORM,
    split_by_head_config={0},  # just split layer 0
)
# printer.print(split_first)
less_printer_split.print(split_first)

# %%

# you can also pass a dict of a list of tuples (args for explicit_reduce)
split_complex = configure_transformer(
    c,
    to=To.ATTN_HEAD_MLP_NORM,
    split_by_head_config={
        0: [(1, "my_group"), (0, "my_other_group")],
        1: [(S[:1], "first_part"), (1, "snd")],
    },
)
# printer.print(split_complex)
less_printer_split.print(split_complex)

# %%

# if you'd like, you can also apply head splitting afterward

to_split = configure_transformer(c, to=To.ATTN_HEAD_MLP_NORM)
split_after = split_by_head(to_split, split="full")
# printer.print(split_after)
less_printer_split.print(split_after)

# %% [markdown]

# ## The end
#
# this is end of the tutorial part of this notebook. Below here we just have misc testing.

# %%

pushed = push_down_transformer_weights(c, To.BLOCK)
printer.print(pushed)

# %%

pushed = strip_arr(push_down_transformer_weights(c, To.ATTN_MLP))
printer.print(pushed)

# %%

pushed = push_down_transformer_weights(c, To.ATTN_MLP_NORM)
printer.print(pushed)

# %%

pushed = push_down_transformer_weights(c, To.ATTN_HEAD_MLP_NORM)
printer.print(pushed)

# %%

pushed = push_down_transformer_weights(c, To.ATTN_HEAD_MLP_NORM)
# printer.print(pushed)
split = split_by_head(pushed)
printer.print(split)
# pushed.print()

# %%

pushed = push_down_transformer_weights(c, To.ATTN_MLP_NORM)
pushed = clear_block_module(pushed)
# printer.print(pushed)
flat = flatten_res(pushed)
printer.print(flat)

# rc.PrintOptions.add_nest_default().print(pushed.get_unique("b1"))

# printer_few = rc.PrintOptions(
#     bijection=False, traversal=rc.new_traversal(term_early_at=rc.Matcher("a", "m", "ln", "a.head.on_inp", rc.Matcher.regex("^[am].ln_call$")))
# )

# printer_few.print(pushed)

# %%

pushed = push_down_transformer_weights(pull_up_bias(c), To.ATTN_HEAD_MLP_NORM)
pushed = split_by_head(pushed)
pushed = clear_block_module(pushed)
# pushed = pull_up_head_split(pushed)
printer.print(pushed)
# printer.print(c)

# %%

pushed = push_down_transformer_weights(pull_up_bias(c), To.ATTN_HEAD_MLP_NORM)
pushed = split_by_head(pushed)
pushed = clear_block_module(pushed)
pushed = pull_up_head_split(pushed)
flat = flatten_res(pushed, flatten_components=True)
less_printer_split.print(flat)


# %%

with pytest.raises(AssertionError):
    configure_transformer(c, to=None, use_clear_block_module=True, use_strip_arr=False)

# %%

params_nb = mod_l.TransformerParams(mod_l.TransformerBlockParams(attn_bias=False, mlp_output_bias=False), num_layers=2)
c_nb, _, _ = params_nb.garbage_call(num_heads=3)
c_nb = c_nb.get_unique("t.bind_w")

# %%

yes_pull = configure_transformer(c_nb, use_pull_up_bias=True)
# printer.print(default)
less_printer.print(yes_pull)

# %%

# same as above, no bias
no_pull = configure_transformer(c_nb, use_pull_up_bias=False)
# printer.print(default)
less_printer.print(no_pull)

# %%

x = configure_transformer(
    c_nb,
    to=To.ATTN_HEAD_MLP_NORM,
    split_by_head_config="full",
    use_pull_up_head_split=True,
    use_flatten_res=True,
    flatten_components=True,
)
less_printer_split.print(x)

# %%

params_nb_nn = mod_l.TransformerParams(
    mod_l.TransformerBlockParams(attn_bias=False, mlp_output_bias=False, norm_type=None), num_layers=2
)
c_nb_nn, _, _ = params_nb_nn.garbage_call(num_heads=3)
c_nb_nn = c_nb_nn.get_unique("t.bind_w")

# %%

yes_pull = configure_transformer(c_nb_nn, use_pull_up_bias=True)
# printer.print(default)
less_printer.print(yes_pull)

# %%

# same as above, no bias
no_pull = configure_transformer(c_nb_nn, use_pull_up_bias=False)
# printer.print(default)
less_printer.print(no_pull)


# %%

x = configure_transformer(
    c_nb_nn,
    to=To.ATTN_HEAD_MLP_NORM,
    split_by_head_config="full",
    use_pull_up_head_split=True,
    use_flatten_res=True,
    flatten_components=True,
)
less_printer_split.print(x)

# %%


x = configure_transformer(
    c_nb_nn,
    to=To.ATTN_HEAD_MLP_NORM,
    split_by_head_config="full",
    use_pull_up_head_split=False,
    use_flatten_res=True,
    flatten_components=True,
)
less_printer_split.print(x)

# %%

x = configure_transformer(
    c_nb_nn,
    to=To.ATTN_HEAD_MLP_NORM,
    split_by_head_config="full",
    use_pull_up_head_split=False,
    use_flatten_res=True,
    flatten_components=True,
)
# printer.print(x)
less_printer_split.print(x)



import pytest

from rust_circuit import *

circs = [
    Add(Array.randn(128, 32, 32), Scalar(1.0, (), None)),
    Add(Add(Scalar(5.5, (2, 2), None), Scalar(1.0, (), None)), Scalar(1.0, (), None)),
]


def erroring_function(*args, **kwargs):
    assert False


def test_filter():
    assert list(filter_nodes(circs[0], lambda x: isinstance(x, Scalar))) == [Scalar(1.0, (), None)]
    with pytest.raises(Exception):
        list(filter_nodes(circs[0], erroring_function))


def map_scalarconstant_plus_one(x):
    if isinstance(x, Scalar):
        return Scalar(x.value + 1.0, (), None)
    else:
        return x


def test_deep_map():
    circs[0].print()
    deep_map(circs[0], map_scalarconstant_plus_one).print()
    assert deep_map(circs[0], map_scalarconstant_plus_one) == Add(circs[0].children[0], Scalar(2.0, (), None))
    with pytest.raises(Exception):
        deep_map(circs[0], erroring_function)

    def exact_mapper(x):
        if x == Add(Scalar(5.5, (2, 2), None), Scalar(1.0, (), None)):
            return Scalar(9.9, (2, 2), None)
        else:
            return map_scalarconstant_plus_one(x)

    deep_map(circs[1], exact_mapper).print()
    assert deep_map_preorder(circs[1], exact_mapper) == Add(Scalar(9.9, (2, 2), None), Scalar(2.0, (), None))
    assert deep_map(circs[1], exact_mapper) == deep_map(circs[1], map_scalarconstant_plus_one)


def test_path_get():
    path_get(circs[1], [0, 0]).print()
    assert path_get(circs[1], [0, 0]) == Scalar(5.5, (2, 2), None)


def test_update():
    circs[1].print()
    updated = update_nodes(circs[1], lambda x: x == Scalar(5.5, (2, 2), None), lambda x: Scalar(77.0, (2, 2), None))
    updated.print()
    updated2 = update_path(circs[1], [0, 0], lambda x: Scalar(77.0, (2, 2), None))
    updated2.print()


def test_device_dtype_op_repr_and_conversion():
    tdd = TorchDeviceDtype("cuda:0", "int32")
    assert repr(tdd) == 'TorchDeviceDtype(device="cuda:0", dtype="int32")'

    tdd.op()
    assert repr(tdd.op()) == 'TorchDeviceDtypeOp(device="cuda:0", dtype="int32")'

    assert repr(TorchDeviceDtypeOp()) == "TorchDeviceDtypeOp(device=None, dtype=None)"
    assert repr(TorchDeviceDtypeOp(dtype="int32")) == 'TorchDeviceDtypeOp(device=None, dtype="int32")'
    assert repr(TorchDeviceDtypeOp(device="cpu")) == 'TorchDeviceDtypeOp(device="cpu", dtype=None)'


if __name__ == "__main__":
    test_filter()
    test_deep_map()
    test_path_get()
    test_update()



import os

WORKING_RRFS_DIR = "/working/rrfs"
HOME_RRFS_DIR = f'{os.environ.get("HOME", "/home/ubuntu")}/rrfs'
RRFS_DIR = os.environ.get(
    "RR_RRFS_DIR",
    HOME_RRFS_DIR if os.path.exists(HOME_RRFS_DIR) else WORKING_RRFS_DIR,
)



from typing import List, Optional

import torch

import rust_circuit as rc
from rust_circuit.causal_scrubbing.dataset import Dataset
from rust_circuit.causal_scrubbing.hypothesis import CondSampler, Correspondence, FuncSampler

from .setup import ParenDataset


def check_correspondence_equality(corr_a: Correspondence, corr_b: Correspondence, ds: Dataset, circ: rc.Circuit):
    """
    corr_a is user provided, the one to be checked
    corr_b is the one we are checking equality to

    Details:
        - Assumes matchers were designed for the same circuit (with same names, etc)
        - Checks type matches for all condsamplers. Also checks that FuncSamplers define same equiv classes.
        Wouldn't properly check e.g. FixedUncondSamplers
    """

    # TODO: be less strict about name equality
    check_name_equality(corr_a, corr_b)
    for node_b, matcher_b in corr_b.in_dfs_order():
        node_a = corr_a.get_by_name(node_b.name)
        matcher_a = corr_a.corr[node_a]

        check_cond_sampler_equality(node_a.cond_sampler, node_b.cond_sampler, ds, f"{node_b.name} cond sampler")
        check_cond_sampler_equality(
            node_a.other_inputs_sampler, node_b.other_inputs_sampler, ds, f"{node_b.name} other inputs sampler"
        )
        check_matcher_equality(matcher_a, matcher_b, circ, f"matcher for {node_b.name}")


def check_name_equality(corr_a: Correspondence, corr_b: Correspondence):
    a_names = set(corr_a.i_names.keys())
    b_names = set(corr_b.i_names.keys())

    if a_names != b_names:
        print("Interp node names don't match!")
        print("Missing names:", b_names - a_names)
        print("Extra names:", b_names - a_names)
        assert False


def equiv_classes(arr) -> set[tuple]:
    arr = arr[:, None] if arr.ndim == 1 else arr
    eq_arrs: List[torch.Tensor] = [torch.all(arr == val, dim=1) for val in torch.unique(arr, dim=0)]
    return set((tuple(arr.tolist()) for arr in eq_arrs))


def check_cond_sampler_equality(cs_a: CondSampler, cs_b: CondSampler, ds: Dataset, name: str):
    if cs_a is cs_b:
        return
    assert type(cs_a) == type(cs_b), f"{name} is the wrong type"
    if isinstance(cs_b, FuncSampler):
        assert isinstance(cs_a, FuncSampler)
        assert equiv_classes(cs_a.func(ds)) == equiv_classes(cs_b.func(ds)), (
            f"{name} doesn't match equivilance classes",
            cs_a.func(ds),
            cs_b.func(ds),
        )


def check_matcher_equality(
    matcher_a: rc.IterativeMatcher, matcher_b: rc.IterativeMatcher, circ: rc.Circuit, name: Optional[str] = None
):
    if matcher_a == matcher_b:
        return
    name = "matcher" if name is None else name
    assert circ.get(matcher_a) == circ.get(matcher_b), (
        f"Circuit gotten by {name} matches incorrect node(s)",
        matcher_a,
        matcher_b,
        circ.get(matcher_a),
        circ.get(matcher_b),
    )
    circ_a = circ
    circ_b = circ
    for match in circ.get(matcher_b):
        symbol = rc.Symbol.new_with_random_uuid(shape=match.shape)
        circ_a = circ_a.update(match, lambda _: symbol)
        circ_b = circ_b.update(match, lambda _: symbol)
    assert circ_a == circ_b, (
        f"After replacements for {name}, the circuits don't match. Probably the paths by which they are allowed to match differ.",
        matcher_a,
        matcher_b,
        circ.get(circ_a),
        circ.get(circ_b),
    )


def evaluate_on_dataset(c: rc.Circuit, ds: ParenDataset):
    group = rc.DiscreteVar.uniform_probs_and_group(len(ds))
    transform = rc.Sampler(rc.RunDiscreteVarAllSpec([group]))
    c = c.update("tokens", lambda _: rc.DiscreteVar(ds.tokens, probs_and_group=group))
    c = rc.substitute_all_modules(c)
    return transform.sample(c).evaluate().to(device="cpu", dtype=torch.float32)


def check_circuit_equality(circuit_a: rc.Circuit, circuit_b: rc.Circuit, strict=True):
    from . import causal_scrubbing_experiments as cse

    test_ds = cse.ds[:1000]
    eval_a = evaluate_on_dataset(circuit_a, test_ds)
    eval_b = evaluate_on_dataset(circuit_b, test_ds)
    if not torch.allclose(eval_a, eval_b):
        err = (eval_a - eval_b).abs().mean().item()
        assert False, (
            f"circuits do not evaluate to the same thing. Mean abs error {err}",
            circuit_a,
            circuit_b,
            eval_a,
            eval_b,
        )

    if strict:
        assert circuit_a == circuit_b, ("Circuits are not strictly equal. This includes names.", circuit_a, circuit_b)


# prefixed "t_" instead of "test_" since we don't want pytest to collect them.


def t_ex0a_corr(corr):
    from . import causal_scrubbing_experiments as cse

    check_correspondence_equality(corr, cse.corr0a, cse.ds, cse.circuit)


def t_ex0b_corr(corr):
    from . import causal_scrubbing_experiments as cse

    check_correspondence_equality(corr, cse.corr0b, cse.ds, cse.circuit)


def t_m_10(m_10):
    from . import causal_scrubbing_experiments as cse

    check_matcher_equality(m_10, cse.m_10, cse.circuit, "m_10")


def t_m_20(m_20):
    from . import causal_scrubbing_experiments as cse

    check_matcher_equality(m_20, cse.m_20, cse.circuit, "m_20")


def t_m_21(m_21):
    from . import causal_scrubbing_experiments as cse

    check_matcher_equality(m_21, cse.m_21, cse.circuit, "m_21")


def t_count_cond(count_cond):
    from . import causal_scrubbing_experiments as cse

    check_cond_sampler_equality(count_cond, cse.count_cond, cse.ds, "count_cond")


def t_horizon_cond(horizon_cond):
    from . import causal_scrubbing_experiments as cse

    check_cond_sampler_equality(horizon_cond, cse.horizon_cond, cse.ds, "horizon_cond")


def t_ex1_corr(make_ex1_corr, cond: CondSampler):
    from . import causal_scrubbing_experiments as cse

    check_correspondence_equality(make_ex1_corr(cond), cse.make_ex1_corr(cond), cse.ds, cse.circuit)


def t_start_open_cond(start_open_cond):
    from . import causal_scrubbing_experiments as cse

    check_cond_sampler_equality(start_open_cond, cse.start_open_cond, cse.ds, "start_open_cond")


def t_count_open_cond(count_open_cond):
    from . import causal_scrubbing_experiments as cse

    check_cond_sampler_equality(count_open_cond, cse.count_open_cond, cse.ds, "count_open_cond")


def t_ex2_part1_circuit(circuit):
    from . import causal_scrubbing_experiments as cse

    check_circuit_equality(circuit, cse.ex2_part1_circuit)


def t_m_10_p1(m_10_p1):
    from . import causal_scrubbing_experiments as cse

    check_matcher_equality(m_10_p1, cse.m_10_p1, cse.ex2_part1_circuit, "m_10_p1")


def t_m_20_p1(m_20_p1):
    from . import causal_scrubbing_experiments as cse

    check_matcher_equality(m_20_p1, cse.m_20_p1, cse.ex2_part1_circuit, "m_20_p1")


def t_make_ex2_part1_corr(corr):
    from . import causal_scrubbing_experiments as cse

    check_correspondence_equality(corr, cse.make_ex2_part1_corr(), cse.ds, cse.ex2_part1_circuit)


def t_ex2_part2_circuit(circuit):
    from . import causal_scrubbing_experiments as cse

    check_circuit_equality(circuit, cse.ex2_part2_circuit)


def t_ex2_part2_corr(corr):
    from . import causal_scrubbing_experiments as cse

    check_correspondence_equality(corr, cse.make_ex2_part2_corr(), cse.ds, cse.ex2_part2_circuit)


def t_project_into_direction(fn):
    from . import causal_scrubbing_experiments as cse

    h00 = cse.circuit.get_unique("a0.h0")
    assert fn(h00).name == "a0.h0_projected"
    check_circuit_equality(fn(h00), cse.project_into_direction(h00), strict=False)


def t_ex2_part3_circuit(get_circuit):
    from . import causal_scrubbing_experiments as cse

    check_circuit_equality(get_circuit(cse.ex2_part2_circuit, cse.project_into_direction), cse.ex2_part3_circuit)


def t_ex2_part3_corr(corr):
    from . import causal_scrubbing_experiments as cse

    check_correspondence_equality(corr, cse.make_ex2_part3_corr(), cse.ds, cse.ex2_part3_circuit)


def t_compute_phi_circuit(fn):
    from . import causal_scrubbing_experiments as cse

    tokens = cse.circuit.get_unique("tokens")
    assert fn(tokens).name == "a0.h0_phi"
    check_circuit_equality(fn(tokens), cse.compute_phi_circuit(tokens), strict=False)


def t_ex2_part4_circuit(get_circuit):
    from . import causal_scrubbing_experiments as cse

    check_circuit_equality(get_circuit(cse.ex2_part2_circuit, cse.compute_phi_circuit), cse.ex2_part4_circuit)


def t_ex2_part4_corr(corr):
    from . import causal_scrubbing_experiments as cse

    check_correspondence_equality(corr, cse.make_ex2_part4_corr(), cse.ds, cse.ex2_part4_circuit)


def t_separate_all_seqpos(fn):
    from . import causal_scrubbing_experiments as cse

    start_circ = rc.Array(torch.randn((42, 56)), name="start_arr")
    check_circuit_equality(fn(start_circ), cse.separate_all_seqpos(start_circ))


def t_ex3_circuit(circ):
    from . import causal_scrubbing_experiments as cse

    check_circuit_equality(circ, cse.ex3_circuit)


def t_to_horizon_vals(fn):
    from . import causal_scrubbing_experiments as cse

    for i in (0, 1, 15, 41):  # random seqpos
        for adj in (True, False):
            assert equiv_classes(fn(cse.ds, i, adj)) == equiv_classes(cse.to_horizon_vals(cse.ds, i, adj)), (
                f"Equiv classes don't match for i={i}, adj={adj}",
                fn(cse.ds, i, adj),
                cse.to_horizon_vals(cse.ds, i, adj),
            )


def t_get_horizon_all_cond(fn):
    from . import causal_scrubbing_experiments as cse

    for adj in (True, False):
        check_cond_sampler_equality(fn(adj), cse.get_horizon_all_cond(adj), cse.ds, "horizon_all cond")


def t_make_ex3_corr(fn):
    from . import causal_scrubbing_experiments as cse

    for adj in (True, False):
        check_correspondence_equality(fn(adj), cse.make_ex3_corr(adj), cse.ds, cse.ex3_circuit)


def t_ex4_corr(corr):
    from . import causal_scrubbing_experiments as cse

    check_correspondence_equality(corr, cse.ex4_corr, cse.ds, cse.ex4_circuit)



from rust_circuit import Array, Einsum, Rearrange, SimpFnSubset, simp

# %%

fin = Rearrange.from_string(Einsum.from_einsum_string("abc->cab", Array.randn(2, 3, 4, name="hi")), "c a b -> a c b")
print()
fin.print()

# %%

print()
simp(fin).print()

# %%

print()
SimpFnSubset.compiler_default().simp(fin).print()

# %%

print()
# same as just `simp`
SimpFnSubset.default().simp(fin).print()

# %%

print()
# disable a function
SimpFnSubset.default().set(rearrange_fuse=False).simp(fin).print()

# %%

# fancy repr
assert str(SimpFnSubset.default()) == "SimpFnSubset.default()"
print(str(SimpFnSubset.default().set(rearrange_fuse=False)))
assert (
    str(SimpFnSubset.default().set(rearrange_fuse=False))
    == """SimpFnSubset.default().set(
    # Rearrange
    rearrange_fuse = False,
)"""
)
print(str(SimpFnSubset.default().set(rearrange_fuse=False, einsum_concat_to_add=True)))
assert (
    str(SimpFnSubset.default().set(rearrange_fuse=False, einsum_concat_to_add=True))
    == """SimpFnSubset.default().set(
    # Einsum
    einsum_concat_to_add = True,
    # Rearrange
    rearrange_fuse = False,
)"""
)

# %%

# TODO: add support for partial simp with traversal and demo this!



import warnings
from functools import lru_cache
from typing import Any, Dict, Optional

import attrs
import einops
import numpy as np
import torch
from fancy_einsum import einsum
from scipy.special import softmax
from torch.nn.functional import layer_norm, one_hot

import rust_circuit as rc
from adversarial.simple_task.dataset_utils import load_dataset
from rust_circuit.causal_scrubbing.dataset import Dataset
from rust_circuit.interop_rust import cached_circuit_by_hash
from rust_circuit.module_library import load_model_id

##################  TOKENIZER ##################


class ParenTokenizer:
    START_TOKEN = 0
    PAD_TOKEN = 1
    END_TOKEN = 2
    OPEN_TOKEN = 3
    CLOSE_TOKEN = 4
    vocab_size = 5

    @classmethod
    def tokenize(cls, strs: list[str], max_len: Optional[int] = None) -> torch.Tensor:
        if max_len is None:
            max_len = max((max(len(s) for s in strs), 1))

        tokenizer = rc.CharTokenizer(
            start=cls.START_TOKEN,
            end=cls.END_TOKEN,
            pad=cls.PAD_TOKEN,
            pad_width=max_len + 2,  # this is length of output, so needs space for begin+end
            mapping={"(": cls.OPEN_TOKEN, ")": cls.CLOSE_TOKEN},
            error_if_over=False,
        )

        return tokenizer.tokenize_strings(strs)

    @classmethod
    def decode(cls, tokens: torch.Tensor) -> list[str]:
        def int_to_c(c: float) -> str:
            if c == cls.OPEN_TOKEN:
                return "("
            if c == cls.CLOSE_TOKEN:
                return ")"
            else:
                raise ValueError(c)

        return [
            "".join(int_to_c(i.item()) for i in seq[1:] if i != cls.PAD_TOKEN and i != cls.END_TOKEN) for seq in tokens
        ]


##################  DATA SET ##################


@cached_circuit_by_hash
def compute_p_open_after(circuit: rc.Array) -> torch.Tensor:
    count_open_after = cumsum_reversed(circuit.value == ParenTokenizer.OPEN_TOKEN, 1)
    count_close_after = cumsum_reversed(circuit.value == ParenTokenizer.CLOSE_TOKEN, 1)
    return count_open_after / (count_open_after + count_close_after)


@cached_circuit_by_hash
def compute_adjusted_p_open_after(circuit: rc.Array, model_id: Any) -> torch.Tensor:
    # there are several divide-by-zeros and means of empty slices that result in nan values.
    # that's okay! These nan values shouldn't ever be referenced (except if the dataset has non-empty strs)
    # so we'll just catch the warnings
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        weights = get_adjusted_attn_factors(model_id)
    is_open = circuit.value == ParenTokenizer.OPEN_TOKEN
    # broadcast both to b * q * k
    weighted_opens = weights[(circuit.value != ParenTokenizer.PAD_TOKEN).sum(-1)] * is_open[:, None, :]
    return weighted_opens.sum(-1)


def cumsum_reversed(x: torch.Tensor, dim: int):
    return x.sum(dim, keepdim=True) - x.cumsum(dim) + x


@attrs.frozen
class ParenDataset(Dataset):
    model_id: str = attrs.field()

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

    @property
    def p_open_after(self):
        return compute_p_open_after(self.tokens_flat)

    @property
    def starts_with_open(self) -> torch.Tensor:
        return self.tokens_flat.value[:, 1] == ParenTokenizer.OPEN_TOKEN

    @property
    def count_test(self) -> torch.Tensor:
        return self.p_open_after[:, 0] == 0.5

    @property
    def horizon_test(self) -> torch.Tensor:
        return torch.tensor(np.nanmax(self.p_open_after, axis=1) <= 0.5, dtype=self.tokens.value.dtype)

    @property
    def strs(self):
        return ["b" + s + "e" for s in ParenTokenizer.decode(self.tokens_flat.value)]

    @property
    def input_lengths(self) -> torch.Tensor:
        return (self.tokens_flat.value != ParenTokenizer.PAD_TOKEN).sum(-1)

    @classmethod
    def load(cls, model_id="jun9_paren_balancer", dataset_name="random_choice_len_40_extra_yeses_8"):
        data_list = load_dataset("balanced_parens", dataset_name)["dev"]
        inputs, labels = [s for s, a in data_list], [a for s, a in data_list]

        toks = ParenTokenizer.tokenize(inputs, max_len=40)
        labels = torch.tensor(labels, dtype=torch.float32)

        one_hot_toks = one_hot(toks, ParenTokenizer.vocab_size).to(dtype=torch.float32)

        return cls(
            (rc.Array(toks, name="tokens_flat"), rc.Array(one_hot_toks, name="tokens"), rc.Array(labels, name="is_balanced")),  # type: ignore
            model_id=model_id,
        )

    @property
    def adjusted_p_open_after(self):
        return compute_adjusted_p_open_after(self.tokens_flat, self.model_id)

    def str_values(self) -> str:
        return f"{[('strs', self.strs), (self.is_balanced.name, self.is_balanced.value.to(torch.bool))]}"

    def __repr__(self) -> str:
        return str(self)

    def __str__(self) -> str:
        if len(self) == 1:
            s = "ParenDS("
            s += " str:" + self.strs[0]
            s += " bal:" + ("T" if self.is_balanced.value.item() else "F")
            s += " count:" + ("T" if self.count_test.item() else "F")
            s += " horz:" + ("T" if self.horizon_test.item() else "F")
            s += " len:" + str(self.input_lengths.item()) + ")"
            return s
        else:
            return f"ParenDataset(len={len(self)})"


@lru_cache
def get_weights(model_id: str) -> Dict[str, torch.Tensor]:
    circ_dict, _, _ = load_model_id(model_id)
    # we don't include spec child
    return {
        child.name[:-4]: child.cast_array().value  # strip _arr from end
        for i, child in enumerate(circ_dict["t.bind_w"].children)
        if i != 0  # don't include spec
    } | {
        "tok_embeds": circ_dict["t.w.tok_embeds"].cast_array().value,
        "pos_embeds": circ_dict["t.w.pos_embeds"].cast_array().value,
    }


################## ADJUSTED ATTN FACTOR COMPUTATION ##################
# (embarassingly long code for something so simple, don't worry about it)


@lru_cache
def get_adjusted_attn_factors(model_id: str):
    """Returns shape [inputlen, qpos] of the proportion of 0.0's-attention-to-paren-positions that are on or after qpos
    rather than before it. Ignores attn not on these positions for the proportion."""
    attn_probs = calc_h00_attn_probs(model_id)  # [inputlen, qpos, kpos]

    masks = get_attn_arr_masks()

    def avg_attn_in_mask(mask):
        mean_where = torch.tensor(np.array(attn_probs).mean(axis=-1, keepdims=True, where=np.array(mask)))
        avg_by_key = torch.nan_to_num(mean_where, nan=0)
        return mask * avg_by_key

    weights = avg_attn_in_mask(masks["paren_before_qpos"]) + avg_attn_in_mask(masks["paren_on_or_after_qpos"])
    weights_normed = weights / weights.sum(dim=-1, keepdim=True)

    return torch.tensor(weights_normed)


def a0_layer_norm(x, model_id: str):
    weights = get_weights(model_id)
    return layer_norm(x, (56,), weight=weights["a0.ln.w.scale"], bias=weights["a0.ln.w.bias"])


def make_input_arr(model_id: str, just_opens=False) -> torch.Tensor:
    """
    Returns an array of [input len, seq pos, d model]
    There are 41 possible input lens (from 2 to 42), but we'll make that dimension 43 long so indexing is easy.
    By default averages the ln-ed open and close positions for the positions corresponding to parens. If just_opens is true, it just uses the open embeds
    """
    weights = get_weights(model_id)
    # [pos, tok, d_model] matrix of possible input embeds
    possible_embeds = weights["pos_embeds"][:42, None, :] + weights["tok_embeds"][None, :, :]
    possible_ln_embeds = a0_layer_norm(possible_embeds, model_id)

    if just_opens:
        paren_vals = possible_ln_embeds[:, ParenTokenizer.OPEN_TOKEN, :]
    else:
        paren_vals = possible_ln_embeds[:, [ParenTokenizer.OPEN_TOKEN, ParenTokenizer.CLOSE_TOKEN], :].mean(1)

    input_lens_arr = einops.repeat(torch.arange(43), "inputlens -> inputlens seqpos dmodel", seqpos=42, dmodel=56)
    seq_pos_arr = einops.repeat(torch.arange(42), "seqpos -> inputlens seqpos dmodel", inputlens=43, dmodel=56)

    arr = torch.where(seq_pos_arr < input_lens_arr, paren_vals, possible_ln_embeds[:, ParenTokenizer.PAD_TOKEN, :])
    arr = torch.where(seq_pos_arr == 0, possible_ln_embeds[None, 0, ParenTokenizer.START_TOKEN, :], arr)
    arr = torch.where(seq_pos_arr == input_lens_arr - 1, possible_ln_embeds[None, :, ParenTokenizer.END_TOKEN, :], arr)
    arr[[0, 1], :, :] = torch.nan
    return arr


def get_attn_arr_masks() -> Dict[str, torch.Tensor]:
    """Returns a bunch of helpful masks of shape (43, 42, 42) = (inputpos, qpos, kpos)"""
    inputlens_arr = einops.repeat(torch.arange(43), "inputlens -> inputlens qpos kpos", qpos=42, kpos=42)
    qpos_arr = einops.repeat(torch.arange(42), "qpos -> inputlens qpos kpos", inputlens=43, kpos=42)
    kpos_arr = einops.repeat(torch.arange(42), "kpos -> inputlens qpos kpos", inputlens=43, qpos=42)

    masks = {
        "paren_pos": (kpos_arr >= 1) & (kpos_arr <= inputlens_arr - 2),
        "padding_pos": kpos_arr >= inputlens_arr,
        "kpos_before_qpos": (kpos_arr < qpos_arr),
        "kpos_on_or_after_qpos": (kpos_arr >= qpos_arr),
    }
    masks["paren_before_qpos"] = masks["paren_pos"] & masks["kpos_before_qpos"]
    masks["paren_on_or_after_qpos"] = masks["paren_pos"] & masks["kpos_on_or_after_qpos"]
    return masks


def calc_h00_attn_probs(model_id: str) -> torch.Tensor:
    weights = get_weights(model_id)

    query_inputs = make_input_arr(model_id, just_opens=True)
    key_inputs = make_input_arr(model_id, just_opens=False)

    queries = (
        einsum("dhead dmodel, inplen qpos dmodel -> inplen qpos dhead", weights["a0.w.q"][0], query_inputs)
        + weights["a0.w.q_bias"][0]
    )
    keys = (
        einsum("dhead dmodel, inplen kpos dmodel -> inplen kpos dhead", weights["a0.w.k"][0], key_inputs)
        + weights["a0.w.k_bias"][0]
    )
    unmasked_scores = einsum("inplen qpos dmodel, inplen kpos dmodel -> inplen qpos kpos", queries, keys)

    masks = get_attn_arr_masks()
    masked_scores = torch.where(masks["padding_pos"], unmasked_scores - 10_000, unmasked_scores)
    scores = masked_scores / (28 ** 0.5)
    return softmax(scores, axis=-1)


def get_h00_open_vector(model_id: str) -> torch.Tensor:
    """
    This function estimates a single vector v which head 0.0 would theoretically output if p=1 (the entire sequence is open parens). It would thus (hypothetically) output -v if p=0, and 0 if p=0.5.
    """
    weights = get_weights(model_id)

    def through_00(x):
        ln_out = a0_layer_norm(x, model_id)
        v_out = torch.einsum("oi, ...i -> ...o", weights["a0.w.v"][0], ln_out) + weights["a0.w.v_bias"][0]
        return torch.einsum("oi, ...i -> ...o", weights["a0.w.o"][0], v_out)

    open_outs = through_00(weights["tok_embeds"][ParenTokenizer.OPEN_TOKEN] + weights["pos_embeds"][1:41])
    close_outs = through_00(weights["tok_embeds"][ParenTokenizer.CLOSE_TOKEN] + weights["pos_embeds"][1:41])

    # empirically all the open outs are close together, and nearly-opposite all hte close outs. But we need to
    # average them together to return a single vector
    return (open_outs.mean(0) - close_outs.mean(0)) / 2
